@@ -16,6 +16,8 @@ package command
 
 import (
 	"context"
+	"encoding/json"
+	"fmt"
 
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/vulncheck"
@@ -186,16 +188,30 @@ type Interface interface {
 	// runner.
 	StopProfile(context.Context, StopProfileArgs) (StopProfileResult, error)
 
-	// RunGovulncheck: Run vulncheck
+	// GoVulncheck: run vulncheck synchronously.
 	//
 	// Run vulnerability check (`govulncheck`).
 	//
-	// This command is asynchronous; clients must wait for the 'end' progress notification.
+	// This command is synchronous, and returns the govulncheck result.
+	Vulncheck(context.Context, VulncheckArgs) (VulncheckResult, error)
+
+	// RunGovulncheck: Run vulncheck asynchronously.
+	//
+	// Run vulnerability check (`govulncheck`).
+	//
+	// This command is asynchronous; clients must wait for the 'end' progress
+	// notification and then retrieve results using gopls.fetch_vulncheck_result.
+	//
+	// Deprecated: clients should call gopls.vulncheck instead, which returns the
+	// actual vulncheck result.
 	RunGovulncheck(context.Context, VulncheckArgs) (RunVulncheckResult, error)
 
 	// FetchVulncheckResult: Get known vulncheck result
 	//
 	// Fetch the result of latest vulnerability check (`govulncheck`).
+	//
+	// Deprecated: clients should call gopls.vulncheck instead, which returns the
+	// actual vulncheck result.
 	FetchVulncheckResult(context.Context, URIArg) (map[protocol.DocumentURI]*vulncheck.Result, error)
 
 	// MemStats: Fetch memory statistics
@@ -224,6 +240,9 @@ type Interface interface {
 	// to avoid conflicts with other counters gopls collects.
 	AddTelemetryCounters(context.Context, AddTelemetryCountersArgs) error
 
+	// AddTest: add test for the selected function
+	AddTest(context.Context, protocol.Location) (*protocol.WorkspaceEdit, error)
+
 	// MaybePromptForTelemetry: Prompt user to enable telemetry
 	//
 	// Checks for the right conditions, and then prompts the user
@@ -506,20 +525,11 @@ type RunVulncheckResult struct {
 	Token protocol.ProgressToken
 }
 
-// CallStack models a trace of function calls starting
-// with a client function or method and ending with a
-// call to a vulnerable symbol.
-type CallStack []StackEntry
-
-// StackEntry models an element of a call stack.
-type StackEntry struct {
-	// See golang.org/x/exp/vulncheck.StackEntry.
-
-	// User-friendly representation of function/method names.
-	// e.g. package.funcName, package.(recvType).methodName, ...
-	Name string
-	URI  protocol.DocumentURI
-	Pos  protocol.Position // Start position. (0-based. Column is always 0)
+// GovulncheckResult holds the result of synchronously running the vulncheck
+// command.
+type VulncheckResult struct {
+	// Result holds the result of running vulncheck.
+	Result *vulncheck.Result
 }
 
 // MemStatsResult holds selected fields from runtime.MemStats.
@@ -574,12 +584,68 @@ type AddTelemetryCountersArgs struct {
 }
 
 // ChangeSignatureArgs specifies a "change signature" refactoring to perform.
+//
+// The new signature is expressed via the NewParams and NewResults fields. The
+// elements of these lists each describe a new field of the signature, by
+// either referencing a field in the old signature or by defining a new field:
+//   - If the element is an integer, it references a positional parameter in the
+//     old signature.
+//   - If the element is a string, it is parsed as a new field to add.
+//
+// Suppose we have a function `F(a, b int) (string, error)`. Here are some
+// examples of refactoring this signature in practice, eliding the 'Location'
+// and 'ResolveEdits' fields.
+//   - `{ "NewParams": [0], "NewResults": [0, 1] }` removes the second parameter
+//   - `{ "NewParams": [1, 0], "NewResults": [0, 1] }` flips the parameter order
+//   - `{ "NewParams": [0, 1, "a int"], "NewResults": [0, 1] }` adds a new field
+//   - `{ "NewParams": [1, 2], "NewResults": [1] }` drops the `error` result
 type ChangeSignatureArgs struct {
-	RemoveParameter protocol.Location
+	// Location is any range inside the function signature. By convention, this
+	// is the same location provided in the codeAction request.
+	Location protocol.Location // a range inside of the function signature, as passed to CodeAction
+
+	// NewParams describes parameters of the new signature.
+	// An int value references a parameter in the old signature by index.
+	// A string value describes a new parameter field (e.g. "x int").
+	NewParams []ChangeSignatureParam
+
+	// NewResults describes results of the new signature (see above).
+	// An int value references a result in the old signature by index.
+	// A string value describes a new result field (e.g. "err error").
+	NewResults []ChangeSignatureParam
+
 	// Whether to resolve and return the edits.
 	ResolveEdits bool
 }
 
+// ChangeSignatureParam implements the API described in the doc string of
+// [ChangeSignatureArgs]: a union of JSON int | string.
+type ChangeSignatureParam struct {
+	OldIndex int
+	NewField string
+}
+
+func (a *ChangeSignatureParam) UnmarshalJSON(b []byte) error {
+	var s string
+	if err := json.Unmarshal(b, &s); err == nil {
+		a.NewField = s
+		return nil
+	}
+	var i int
+	if err := json.Unmarshal(b, &i); err == nil {
+		a.OldIndex = i
+		return nil
+	}
+	return fmt.Errorf("must be int or string")
+}
+
+func (a ChangeSignatureParam) MarshalJSON() ([]byte, error) {
+	if a.NewField != "" {
+		return json.Marshal(a.NewField)
+	}
+	return json.Marshal(a.OldIndex)
+}
+
 // DiagnoseFilesArgs specifies a set of files for which diagnostics are wanted.
 type DiagnoseFilesArgs struct {
 	Files []protocol.DocumentURI
